from .boundingbox import BoxFormat, Coordinates, BoundingBox
from .annotation import Annotation
from .annotationset import AnnotationSet
from .errors import UnknownImageFormat, ParsingError, FileParsingError
from .evalutation import EvaluationItem, Evaluation, COCOEvaluator